


class DigestingError(Exception):
    pass


class UnsupportedIndicatorError(Exception):
    pass


class RecoveringError(Exception):
    """Wasn't able to recover an object from dictionary using indication."""
